@@ -7,11 +7,13 @@ import os
 import pwd
 import sys
 from contextlib import suppress
-from typing import NamedTuple, Optional, Set
+from typing import NamedTuple, Optional, Set, TYPE_CHECKING
 
-from .options_stub import Options
 from .types import run_once
 
+if TYPE_CHECKING:
+    from .options.types import Options
+
 
 class Version(NamedTuple):
     major: int
@@ -21,7 +23,7 @@ class Version(NamedTuple):
 
 appname: str = 'kitty'
 kitty_face = '🐱'
-version: Version = Version(0, 20, 1)
+version: Version = Version(0, 21, 0)
 str_version: str = '.'.join(map(str, version))
 _plat = sys.platform.lower()
 is_macos: bool = 'darwin' in _plat
@@ -149,7 +151,7 @@ def detect_if_wayland_ok() -> bool:
     return True
 
 
-def is_wayland(opts: Optional[Options] = None) -> bool:
+def is_wayland(opts: Optional['Options'] = None) -> bool:
     if is_macos:
         return False
     if opts is None:
