@@ -7,7 +7,7 @@ import os
 import shutil
 import sys
 from contextlib import contextmanager, suppress
-from typing import Dict, Generator, List, Optional, Sequence
+from typing import Dict, Generator, List, Optional, Sequence, Tuple
 
 from .borders import load_borders_program
 from .boss import Boss
@@ -21,13 +21,13 @@ from .constants import (
     is_wayland, kitty_exe, logo_png_file, running_in_kitty
 )
 from .fast_data_types import (
-    GLFW_IBEAM_CURSOR, create_os_window, free_font_data, glfw_init,
-    glfw_terminate, load_png_data, set_custom_cursor, set_default_window_icon,
-    set_options
+    GLFW_IBEAM_CURSOR, GLFW_MOD_ALT, GLFW_MOD_SHIFT, create_os_window,
+    free_font_data, glfw_init, glfw_terminate, load_png_data,
+    set_custom_cursor, set_default_window_icon, set_options
 )
 from .fonts.box_drawing import set_scale
 from .fonts.render import set_font_family
-from .options_stub import Options as OptionsStub
+from .options.types import Options
 from .os_window_size import initial_window_size_func
 from .session import get_os_window_sizing_data
 from .types import SingleKey
@@ -98,24 +98,29 @@ def init_glfw_module(glfw_module: str, debug_keyboard: bool = False, debug_rende
         raise SystemExit('GLFW initialization failed')
 
 
-def init_glfw(opts: OptionsStub, debug_keyboard: bool = False, debug_rendering: bool = False) -> str:
+def init_glfw(opts: Options, debug_keyboard: bool = False, debug_rendering: bool = False) -> str:
     glfw_module = 'cocoa' if is_macos else ('wayland' if is_wayland(opts) else 'x11')
     init_glfw_module(glfw_module, debug_keyboard, debug_rendering)
     return glfw_module
 
 
-def get_macos_shortcut_for(opts: OptionsStub, function: str = 'new_os_window') -> Optional[SingleKey]:
+def get_macos_shortcut_for(opts: Options, function: str = 'new_os_window', args: Tuple = (), lookup_name: str = '') -> Optional[SingleKey]:
     ans = None
     candidates = []
     for k, v in opts.keymap.items():
-        if v.func == function:
+        if v.func == function and v.args == args:
             candidates.append(k)
     if candidates:
         from .fast_data_types import cocoa_set_global_shortcut
-
+        alt_mods = GLFW_MOD_ALT, GLFW_MOD_ALT | GLFW_MOD_SHIFT
         # Reverse list so that later defined keyboard shortcuts take priority over earlier defined ones
         for candidate in reversed(candidates):
-            if cocoa_set_global_shortcut(function, candidate[0], candidate[2]):
+            if candidate.mods in alt_mods:
+                # Option based shortcuts dont work in the global menubar,
+                # presumably because Apple reserves them for IME, see
+                # https://github.com/kovidgoyal/kitty/issues/3515
+                continue
+            if cocoa_set_global_shortcut(lookup_name or function, candidate[0], candidate[2]):
                 ans = candidate
                 break
     return ans
@@ -127,13 +132,17 @@ def set_x11_window_icon() -> None:
     set_default_window_icon(path + '-128' + ext)
 
 
-def _run_app(opts: OptionsStub, args: CLIOptions, bad_lines: Sequence[BadLine] = ()) -> None:
+def _run_app(opts: Options, args: CLIOptions, bad_lines: Sequence[BadLine] = ()) -> None:
     global_shortcuts: Dict[str, SingleKey] = {}
     if is_macos:
-        for ac in ('new_os_window', 'close_os_window', 'close_tab', 'edit_config_file', 'previous_tab', 'next_tab', 'new_tab'):
+        for ac in ('new_os_window', 'close_os_window', 'close_tab', 'edit_config_file', 'previous_tab',
+                   'next_tab', 'new_tab', 'new_window', 'close_window'):
             val = get_macos_shortcut_for(opts, ac)
             if val is not None:
                 global_shortcuts[ac] = val
+        val = get_macos_shortcut_for(opts, 'clear_terminal', args=('reset', True), lookup_name='reset_terminal')
+        if val is not None:
+            global_shortcuts['reset_terminal'] = val
     if is_macos and opts.macos_custom_beam_cursor:
         set_custom_ibeam_cursor()
     if not is_wayland() and not is_macos:  # no window icons on wayland
@@ -163,14 +172,18 @@ class AppRunner:
         self.first_window_callback = lambda window_handle: None
         self.initial_window_size_func = initial_window_size_func
 
-    def __call__(self, opts: OptionsStub, args: CLIOptions, bad_lines: Sequence[BadLine] = ()) -> None:
+    def __call__(self, opts: Options, args: CLIOptions, bad_lines: Sequence[BadLine] = ()) -> None:
         set_scale(opts.box_drawing_scale)
         set_options(opts, is_wayland(), args.debug_rendering, args.debug_font_fallback)
-        set_font_family(opts, debug_font_matching=args.debug_font_fallback)
         try:
+            set_font_family(opts, debug_font_matching=args.debug_font_fallback)
             _run_app(opts, args, bad_lines)
         finally:
+            set_options(None)
             free_font_data()  # must free font data before glfw/freetype/fontconfig/opengl etc are finalized
+            if is_macos:
+                from kitty.fast_data_types import cocoa_set_notification_activated_callback
+                cocoa_set_notification_activated_callback(None)
 
 
 run_app = AppRunner()
@@ -241,7 +254,7 @@ def expand_listen_on(listen_on: str, from_config_file: bool) -> str:
     return listen_on
 
 
-def setup_environment(opts: OptionsStub, cli_opts: CLIOptions) -> None:
+def setup_environment(opts: Options, cli_opts: CLIOptions) -> None:
     from_config_file = False
     if not cli_opts.listen_on and opts.listen_on.startswith('unix:'):
         cli_opts.listen_on = opts.listen_on
@@ -297,10 +310,10 @@ def _main() -> None:
         os.chdir(os.path.expanduser('~'))
     cli_opts, rest = parse_args(args=args, result_class=CLIOptions)
     cli_opts.args = rest
-    if cli_opts.debug_config:
-        create_opts(cli_opts, debug_config=True)
-        return
     if cli_opts.detach:
+        if cli_opts.session == '-':
+            from .session import PreReadSession
+            cli_opts.session = PreReadSession(sys.stdin.read())
         detach()
     if cli_opts.replay_commands:
         from kitty.client import main as client_main
