@@ -19,29 +19,30 @@ from typing import (
 from .child import ProcessDesc
 from .cli_stub import CLIOptions
 from .config import build_ansi_color_table
-from .constants import appname, wakeup, is_macos
+from .constants import appname, is_macos, wakeup
 from .fast_data_types import (
     BGIMAGE_PROGRAM, BLIT_PROGRAM, CELL_BG_PROGRAM, CELL_FG_PROGRAM,
-    CELL_PROGRAM, CELL_SPECIAL_PROGRAM, DCS, DECORATION, DIM, GLFW_MOD_CONTROL,
+    CELL_PROGRAM, CELL_SPECIAL_PROGRAM, CURSOR_BEAM, CURSOR_BLOCK,
+    CURSOR_UNDERLINE, DCS, DECORATION, DIM, GLFW_MOD_CONTROL,
     GRAPHICS_ALPHA_MASK_PROGRAM, GRAPHICS_PREMULT_PROGRAM, GRAPHICS_PROGRAM,
-    MARK, MARK_MASK, OSC, REVERSE, SCROLL_FULL, SCROLL_LINE, SCROLL_PAGE,
-    STRIKETHROUGH, TINT_PROGRAM, KeyEvent, Screen, add_timer, add_window,
-    cell_size_for_window, compile_program, encode_key_for_tty, get_boss,
-    get_clipboard_string, init_cell_program, pt_to_px, set_clipboard_string,
+    MARK, MARK_MASK, NO_CURSOR_SHAPE, OSC, REVERSE, SCROLL_FULL, SCROLL_LINE,
+    SCROLL_PAGE, STRIKETHROUGH, TINT_PROGRAM, KeyEvent, Screen, add_timer,
+    add_window, cell_size_for_window, click_mouse_url, compile_program,
+    encode_key_for_tty, get_boss, get_clipboard_string, get_options,
+    init_cell_program, mouse_selection, pt_to_px, set_clipboard_string,
     set_titlebar_color, set_window_padding, set_window_render_data,
     update_window_title, update_window_visibility, viewport_for_window
 )
 from .keys import keyboard_mode_name
 from .notify import NotificationCommand, handle_notification_cmd
-from .options_stub import Options
+from .options.types import Options
 from .rgb import to_color
 from .terminfo import get_capabilities
-from .types import ScreenGeometry, WindowGeometry
+from .types import MouseEvent, ScreenGeometry, WindowGeometry
 from .typing import BossType, ChildType, EdgeLiteral, TabType, TypedDict
 from .utils import (
     color_as_int, get_primary_selection, load_shaders, log_error, open_cmd,
-    open_url, parse_color_set, read_shell_environment, sanitize_title,
-    set_primary_selection
+    open_url, parse_color_set, sanitize_title, set_primary_selection
 )
 
 MatchPatternType = Union[Pattern[str], Tuple[Pattern[str], Optional[Pattern[str]]]]
@@ -151,7 +152,8 @@ def as_text(
     as_ansi: bool = False,
     add_history: bool = False,
     add_wrap_markers: bool = False,
-    alternate_screen: bool = False
+    alternate_screen: bool = False,
+    add_cursor: bool = False
 ) -> str:
     lines: List[str] = []
     add_history = add_history and not (screen.is_using_alternate_linebuf() ^ alternate_screen)
@@ -160,6 +162,19 @@ def as_text(
     else:
         f = screen.as_text_non_visual if add_history else screen.as_text
     f(lines.append, as_ansi, add_wrap_markers)
+    ctext = ''
+    if add_cursor:
+        ctext += '\x1b[?25' + ('h' if screen.cursor_visible else 'l')
+        ctext += f'\x1b[{screen.cursor.y + 1};{screen.cursor.x + 1}H'
+        shape = screen.cursor.shape
+        if shape == NO_CURSOR_SHAPE:
+            ctext += '\x1b[?12' + ('h' if screen.cursor.blink else 'l')
+        else:
+            code = {CURSOR_BLOCK: 1, CURSOR_UNDERLINE: 3, CURSOR_BEAM: 5}[shape]
+            if not screen.cursor.blink:
+                code += 1
+            ctext += f'\x1b[{code} q'
+
     if add_history:
         h: List[str] = []
         pht = screen.historybuf.pagerhist_as_text()
@@ -174,8 +189,14 @@ def as_text(
                 h[-1] += '\n'
             if as_ansi:
                 h[-1] += '\x1b[m'
-        return ''.join(chain(h, lines))
-    return ''.join(lines)
+        ans = ''.join(chain(h, lines))
+        if ctext:
+            ans += ctext
+        return ans
+    ans = ''.join(lines)
+    if ctext:
+        ans += ctext
+    return ans
 
 
 class LoadShaderPrograms:
@@ -289,13 +310,13 @@ class Window:
         self,
         tab: TabType,
         child: ChildType,
-        opts: Options,
         args: CLIOptions,
         override_title: Optional[str] = None,
         copy_colors_from: Optional['Window'] = None,
         watchers: Optional[Watchers] = None
     ):
         self.watchers = watchers or Watchers()
+        self.current_mouse_event_button = 0
         self.prev_osc99_cmd = NotificationCommand()
         self.action_on_close: Optional[Callable] = None
         self.action_on_removal: Optional[Callable] = None
@@ -321,8 +342,9 @@ class Window:
         self.geometry: WindowGeometry = WindowGeometry(0, 0, 0, 0, 0, 0)
         self.needs_layout = True
         self.is_visible_in_layout: bool = True
-        self.child, self.opts = child, opts
+        self.child = child
         cell_width, cell_height = cell_size_for_window(self.os_window_id)
+        opts = get_options()
         self.screen: Screen = Screen(self, 24, 80, opts.scrollback_lines, cell_width, cell_height, self.id)
         if copy_colors_from is not None:
             self.screen.copy_colors_from(copy_colors_from.screen)
@@ -341,18 +363,19 @@ class Window:
         q = getattr(self.margin, edge)
         if q is not None:
             return pt_to_px(q, self.os_window_id)
+        opts = get_options()
         if is_single_window:
-            q = getattr(self.opts.single_window_margin_width, edge)
+            q = getattr(opts.single_window_margin_width, edge)
             if q > -0.1:
                 return pt_to_px(q, self.os_window_id)
-        q = getattr(self.opts.window_margin_width, edge)
+        q = getattr(opts.window_margin_width, edge)
         return pt_to_px(q, self.os_window_id)
 
     def effective_padding(self, edge: EdgeLiteral) -> int:
         q = getattr(self.padding, edge)
         if q is not None:
             return pt_to_px(q, self.os_window_id)
-        q = getattr(self.opts.window_padding_width, edge)
+        q = getattr(get_options().window_padding_width, edge)
         return pt_to_px(q, self.os_window_id)
 
     def update_effective_padding(self) -> None:
@@ -368,13 +391,19 @@ class Window:
             self.update_effective_padding()
 
     def effective_border(self) -> int:
-        val, unit = self.opts.window_border_width
+        val, unit = get_options().window_border_width
         if unit == 'pt':
             val = max(1 if val > 0 else 0, pt_to_px(val, self.os_window_id))
         else:
             val = round(val)
         return int(val)
 
+    def apply_options(self) -> None:
+        opts = get_options()
+        self.update_effective_padding()
+        self.change_titlebar_color()
+        setup_colors(self.screen, opts)
+
     @property
     def title(self) -> str:
         return self.override_title or self.child_title
@@ -530,9 +559,18 @@ class Window:
     def use_utf8(self, on: bool) -> None:
         get_boss().child_monitor.set_iutf8_winid(self.id, on)
 
+    def on_mouse_event(self, event: Dict[str, Any]) -> bool:
+        ev = MouseEvent(**event)
+        self.current_mouse_event_button = ev.button
+        action = get_options().mousemap.get(ev)
+        if action is None:
+            return False
+        return get_boss().dispatch_action(action, window_for_dispatch=self, dispatch_type='MouseEvent')
+
     def open_url(self, url: str, hyperlink_id: int, cwd: Optional[str] = None) -> None:
+        opts = get_options()
         if hyperlink_id:
-            if not self.opts.allow_hyperlinks:
+            if not opts.allow_hyperlinks:
                 return
             from urllib.parse import unquote, urlparse, urlunparse
             try:
@@ -551,7 +589,7 @@ class Window:
                         self.handle_remote_file(purl.netloc, unquote(purl.path))
                         return
                     url = urlunparse(purl._replace(netloc=''))
-            if self.opts.allow_hyperlinks & 0b10:
+            if opts.allow_hyperlinks & 0b10:
                 from kittens.tui.operations import styled
                 get_boss()._run_kitten('ask', ['--type=choices', '--message', _(
                     'What would you like to do with this URL:\n') +
@@ -612,13 +650,18 @@ class Window:
     def has_activity_since_last_focus(self) -> bool:
         return self.screen.has_activity_since_last_focus()
 
+    def on_activity_since_last_focus(self) -> None:
+        if get_options().tab_activity_symbol:
+            get_boss().on_activity_since_last_focus(self)
+
     def on_bell(self) -> None:
-        if self.opts.command_on_bell and self.opts.command_on_bell != ['none']:
+        cb = get_options().command_on_bell
+        if cb and cb != ['none']:
             import shlex
             import subprocess
             env = self.child.final_env
             env['KITTY_CHILD_CMDLINE'] = ' '.join(map(shlex.quote, self.child.cmdline))
-            subprocess.Popen(self.opts.command_on_bell, env=env, cwd=self.child.foreground_cwd)
+            subprocess.Popen(cb, env=env, cwd=self.child.foreground_cwd)
         if not self.is_active:
             changed = not self.needs_attention
             self.needs_attention = True
@@ -629,7 +672,8 @@ class Window:
                 tab.on_bell(self)
 
     def change_titlebar_color(self) -> None:
-        val = self.opts.macos_titlebar_color if is_macos else self.opts.wayland_titlebar_color
+        opts = get_options()
+        val = opts.macos_titlebar_color if is_macos else opts.wayland_titlebar_color
         if val:
             if (val & 0xff) == 1:
                 val = self.screen.color_profile.default_bg
@@ -713,7 +757,7 @@ class Window:
             self.refresh()
 
     def request_capabilities(self, q: str) -> None:
-        for result in get_capabilities(q, self.opts):
+        for result in get_capabilities(q, get_options()):
             self.screen.send_escape_code_to_child(DCS, result)
 
     def handle_remote_cmd(self, cmd: str) -> None:
@@ -732,13 +776,14 @@ class Window:
         where, text = data.partition(';')[::2]
         if not where:
             where = 's0'
+        cc = get_options().clipboard_control
         if text == '?':
             response = None
             if 's' in where or 'c' in where:
-                response = get_clipboard_string() if 'read-clipboard' in self.opts.clipboard_control else ''
+                response = get_clipboard_string() if 'read-clipboard' in cc else ''
                 loc = 'c'
             elif 'p' in where:
-                response = get_primary_selection() if 'read-primary' in self.opts.clipboard_control else ''
+                response = get_primary_selection() if 'read-primary' in cc else ''
                 loc = 'p'
             response = response or ''
             from base64 import standard_b64encode
@@ -754,7 +799,7 @@ class Window:
 
             def write(key: str, func: Callable[[str], None]) -> None:
                 if text:
-                    if ('no-append' in self.opts.clipboard_control or
+                    if ('no-append' in cc or
                             len(self.clipboard_control_buffers[key]) > 1024*1024):
                         self.clipboard_control_buffers[key] = ''
                     self.clipboard_control_buffers[key] += text
@@ -763,13 +808,10 @@ class Window:
                 func(self.clipboard_control_buffers[key])
 
             if 's' in where or 'c' in where:
-                if 'write-clipboard' in self.opts.clipboard_control:
+                if 'write-clipboard' in cc:
                     write('c', set_clipboard_string)
             if 'p' in where:
-                if self.opts.copy_on_select == 'clipboard':
-                    if 'write-clipboard' in self.opts.clipboard_control:
-                        write('c', set_clipboard_string)
-                if 'write-primary' in self.opts.clipboard_control:
+                if 'write-primary' in cc:
                     write('p', set_primary_selection)
 
     def manipulate_title_stack(self, pop: bool, title: str, icon: Any) -> None:
@@ -783,10 +825,36 @@ class Window:
                     self.title_stack.append(self.child_title)
     # }}}
 
+    # mouse actions {{{
+    def mouse_click_url(self) -> None:
+        click_mouse_url(self.os_window_id, self.tab_id, self.id)
+
+    def mouse_discard_event(self) -> None:
+        pass
+
+    def mouse_click_url_or_select(self) -> None:
+        if not self.screen.has_selection():
+            self.mouse_click_url()
+
+    def mouse_selection(self, code: int) -> None:
+        mouse_selection(self.os_window_id, self.tab_id, self.id, code, self.current_mouse_event_button)
+
+    def paste_selection(self) -> None:
+        txt = get_boss().current_primary_selection()
+        if txt:
+            self.paste(txt)
+
+    def paste_selection_or_clipboard(self) -> None:
+        txt = get_boss().current_primary_selection_or_clipboard()
+        if txt:
+            self.paste(txt)
+    # }}}
+
     def text_for_selection(self) -> str:
         lines = self.screen.text_for_selection()
-        if self.opts.strip_trailing_spaces == 'always' or (
-                self.opts.strip_trailing_spaces == 'smart' and not self.screen.is_rectangle_select()):
+        sts = get_options().strip_trailing_spaces
+        if sts == 'always' or (
+                sts == 'smart' and not self.screen.is_rectangle_select()):
             return ''.join((ln.rstrip() or '\n') for ln in lines)
         return ''.join(lines)
 
@@ -812,9 +880,10 @@ class Window:
         as_ansi: bool = False,
         add_history: bool = False,
         add_wrap_markers: bool = False,
-        alternate_screen: bool = False
+        alternate_screen: bool = False,
+        add_cursor: bool = False
     ) -> str:
-        return as_text(self.screen, as_ansi, add_history, add_wrap_markers, alternate_screen)
+        return as_text(self.screen, as_ansi, add_history, add_wrap_markers, alternate_screen, add_cursor)
 
     @property
     def cwd_of_child(self) -> Optional[str]:
@@ -841,27 +910,7 @@ class Window:
     def show_scrollback(self) -> None:
         text = self.as_text(as_ansi=True, add_history=True, add_wrap_markers=True)
         data = self.pipe_data(text, has_wrap_markers=True)
-
-        def prepare_arg(x: str) -> str:
-            x = x.replace('INPUT_LINE_NUMBER', str(data['input_line_number']))
-            x = x.replace('CURSOR_LINE', str(data['cursor_y']))
-            x = x.replace('CURSOR_COLUMN', str(data['cursor_x']))
-            return x
-
-        cmd = list(map(prepare_arg, self.opts.scrollback_pager))
-        if not os.path.isabs(cmd[0]):
-            import shutil
-            exe = shutil.which(cmd[0])
-            if not exe:
-                env = read_shell_environment(self.opts)
-                if env and 'PATH' in env:
-                    exe = shutil.which(cmd[0], path=env['PATH'])
-                    if exe:
-                        cmd[0] = exe
-        bdata: Union[str, bytes, None] = data['text']
-        if isinstance(bdata, str):
-            bdata = bdata.encode('utf-8')
-        get_boss().display_scrollback(self, bdata, cmd)
+        get_boss().display_scrollback(self, data['text'], data['input_line_number'])
 
     def paste_bytes(self, text: Union[str, bytes]) -> None:
         # paste raw bytes without any processing
@@ -952,8 +1001,8 @@ class Window:
         self.current_marker_spec = key
 
     def set_marker(self, spec: Union[str, Sequence[str]]) -> None:
-        from .config import parse_marker_spec, toggle_marker
         from .marks import marker_from_spec
+        from .options.utils import parse_marker_spec, toggle_marker
         if isinstance(spec, str):
             func, (ftype, spec_, flags) = toggle_marker('toggle_marker', spec)
         else:
