@@ -91,10 +91,10 @@ class TestKeys(BaseTest):
         mkp('PAGE_DOWN', csi_num=6, trailer='~')
         mkp('HOME', csi_num=1, trailer='H')
         mkp('END', csi_num=1, trailer='F')
-        mods_test(defines.GLFW_FKEY_F1, csi_num=1, trailer='P')
-        mods_test(defines.GLFW_FKEY_F2, csi_num=1, trailer='Q')
-        mods_test(defines.GLFW_FKEY_F3, csi_num=1, trailer='R')
-        mods_test(defines.GLFW_FKEY_F4, csi_num=1, trailer='S')
+        mods_test(defines.GLFW_FKEY_F1, '\x1bOP', csi_num=1, trailer='P')
+        mods_test(defines.GLFW_FKEY_F2, '\x1bOQ', csi_num=1, trailer='Q')
+        mods_test(defines.GLFW_FKEY_F3, '\x1bOR', csi_num=1, trailer='R')
+        mods_test(defines.GLFW_FKEY_F4, '\x1bOS', csi_num=1, trailer='S')
         mods_test(defines.GLFW_FKEY_F5, csi_num=15, trailer='~')
         mods_test(defines.GLFW_FKEY_F6, csi_num=17, trailer='~')
         mods_test(defines.GLFW_FKEY_F7, csi_num=18, trailer='~')
